#include <TLV.h>
#include <stdio.h>

TLVElement :: TLVElement(Int16 type_, Int16 length_)
{
  type   = type_;
  length = length_;
  value  = NULL;
}

TLVElement :: TLVElement(Int16 type_, Int16 length_, byte * value_)
{
  type   = type_;
  length = length_;
  value  = (byte *) Alloc(length);
  memcpy(value, value_, length);
}

TLVElement :: TLVElement(Buffer * buffer)
{
  type   = buffer->GetByte();
  length = buffer->GetByte();
  value  = buffer->GetRaw(length); 
}

TLVElement :: TLVElement(const TLVElement &tlvElement)
{
  type   = tlvElement.type;
  length = tlvElement.length;
  value  = (byte *) Copy(tlvElement.value, length);
}

TLVElement :: ~TLVElement()
{
  type   = 0;
  length = 0;
  Free(value);
  value  = NULL;
}

Int16 TLVElement :: Type() const
{
  return (type);
}

Int16 TLVElement :: Length() const
{
  return (length);
}

byte * TLVElement :: Value() const
{
  return (value);
}

int TLVElement :: Write(Buffer * buffer)
{
  buffer->PutByte(type);
  buffer->PutByte(length);
  buffer->PutRaw(value, length);

  return (1);
}

/****/

TLV :: TLV(bool swap_ = true)
{
  indexmode = false;
  swap      = swap_;
}

TLV :: TLV(Buffer * buffer, bool swap_ = true)
{
  while (buffer->Available() >= 2) {
    TLVElement tlvElement(buffer);
    Append(tlvElement);
  }

  indexmode = false;
  swap      = swap_;
}

TLV :: TLV(Buffer * buffer, int n, bool swap_ = true)
{
  int i;

  for (i = 0; i < n; i ++) {
    TLVElement tlvElement(buffer);
    Append(tlvElement);
  }

  indexmode = false;
  swap      = swap_;
}

TLV :: ~TLV()
{
  indexmode = false;
}

void TLV :: Append(TLVElement &element)
{
  elements.push_back(element);
}

int TLV :: Length()
{
  return (elements.size());
}

int TLV :: Length(Int16 type)
{
  TLVElement * tlvElement = IsPresent(type);

  if (!tlvElement)
    return (-1);

  return (tlvElement->Length());
}

int TLV :: Write(Buffer * buffer)
{
  int i;

  for (i = 0; i < Length(); i ++)
    elements[i].Write(buffer);

  return (1);
}

int TLV :: Read(Buffer * buffer)
{
  while (buffer->Available() > 0) {
    TLVElement tlvElement(buffer);
    Append(tlvElement);
  }

  return (1);
}

int TLV :: Read(Buffer * buffer, int n)
{
  int i;

  for (i = 0; i < n; i ++) {
    TLVElement tlvElement(buffer);
    Append(tlvElement);
  }

  return (1);
}


void TLV :: IndexMode(bool indexmode_)
{
  indexmode = indexmode_;
}

TLVElement * TLV :: IsPresent(Int16 type)
{
  int i;

  if (!indexmode) {
    for (i = 0; i < Length(); i ++)
      if (elements[i].Type() == type)
        return (&elements[i]);
  } else {
    if (type <= Length())
      return (&elements[type]);
  }

  return (NULL);
}

int TLV :: Put(Int16 type)
{
  TLVElement tlvElement(type, 0);

  Append(tlvElement);

  return (1);
}

int TLV :: PutUInt8(Int16 type, UInt8 value)
{
  Int16 length = SizeofInt8;
  byte * data = (byte *) Alloc(length);

  memcpy(data, (byte *) &value, length);

  PutRaw(type, length, data);

  return (1);
}

UInt8 TLV :: GetUInt8(Int16 type)
{
  UInt8 rv;
  Int16 length = SizeofInt8;
  byte * data = GetRaw(type);
  
  if (!data)
    return ((UInt8) -1);
  
  memcpy((byte *) &rv, data, length);
  
  return (rv);
}

int TLV :: PutInt8(Int16 type, Int8 value)
{
  Int16 length = SizeofInt8;
  byte * data = (byte *) Alloc(length);

  memcpy(data, (byte *) &value, length);

  PutRaw(type, length, data);

  return (1);
}

Int8 TLV :: GetInt8(Int16 type)
{
  Int8 rv;
  Int16 length = SizeofInt8;
  byte * data = GetRaw(type);

  if (!data)
    return ((Int8) -1);

  memcpy((byte *) &rv, data, length);

  return (rv);
}

int TLV :: PutUInt16(Int16 type, UInt16 value)
{
  UInt16 ival = swap ? htons(value) : value;
  Int16 length = SizeofInt16;
  byte * data = (byte *) Alloc(length);

  memcpy(data, (byte *) &ival, length);

  PutRaw(type, length, data);

  return (1);
}

UInt16 TLV :: GetUInt16(Int16 type)
{
  UInt16 rv;
  Int16 length = SizeofInt16;
  byte * data = GetRaw(type);

  if (!data)
    return ((UInt16) -1);

  memcpy((byte *) &rv, data, length);

  if (swap)
    rv = ntohs(rv);

  return (rv);
}

int TLV :: PutInt16(Int16 type, Int16 value)
{
  Int16 ival = swap ? htons(value) : value;
  Int16 length = SizeofInt16;
  byte * data = (byte *) Alloc(length);

  memcpy(data, (byte *) &ival, length);

  PutRaw(type, length, data);

  return (1);
}

Int16 TLV :: GetInt16(Int16 type)
{
  Int16 rv;
  Int16 length = SizeofInt16;
  byte * data = GetRaw(type);
  
  if (!data)
    return ((Int16) -1);
  
  memcpy((byte *) &rv, data, length);
  
  if (swap)
    rv = ntohs(rv);
  
  return (rv);
}

int TLV :: PutUInt32(Int16 type, UInt32 value)
{
  UInt32 ival = swap ? htonl(value) : value;
  Int16 length = SizeofInt32;
  byte * data = (byte *) Alloc(length);

  memcpy(data, (byte *) &ival, length);

  PutRaw(type, length, data);

  return (1);
}

UInt32 TLV :: GetUInt32(Int16 type)
{
  UInt32 rv;
  Int16 length = SizeofInt32;
  byte * data = GetRaw(type);

  if (!data)
    return ((UInt32) -1);

  memcpy((byte *) &rv, data, length);

  if (swap)
    rv = ntohl(rv);

  return (rv);
}

int TLV :: PutInt32(Int16 type, Int32 value)
{
  Int32 ival = swap ? htonl(value) : value;
  Int16 length = SizeofInt32;
  byte * data = (byte *) Alloc(length);

  memcpy(data, (byte *) &ival, length);

  PutRaw(type, length, data);

  return (1);
}

Int32 TLV :: GetInt32(Int16 type)
{
  Int32 rv;
  Int16 length = SizeofInt32;
  byte * data = GetRaw(type);

  if (!data)
    return ((Int32) -1);

  memcpy((byte *) &rv, data, length);

  if (swap)
    rv = ntohl(rv);

  return (rv);
}

int TLV :: PutString(Int16 type, string value)
{
  Int16 length = value.length();
  byte * data = (byte *) value.data();

  PutRaw(type, length, data);

  return (1);
}

string TLV :: GetString(Int16 type)
{
  string rv;
  char * value = (char *) GetRaw(type);

  if (value)
    rv.append(value, Length(type));

  return (rv);
}

int TLV :: PutRaw(Int16 type, Int16 length, byte * value)
{
  TLVElement tlvElement(type, length, value);

  Append(tlvElement);

  return (1);
}

byte * TLV :: GetRaw(Int16 type)
{
  byte * rv;
  TLVElement * tlvElement = IsPresent(type);

  if (!tlvElement)
    return (NULL);

  rv = (byte *) Copy(tlvElement->Value(), tlvElement->Length());

  return (rv);
}

int TLV :: PutBuffer(Int16 type, Buffer * buffer)
{
  TLVElement tlvElement(type, buffer->Length(), buffer->DataP());

  Append(tlvElement);

  return (1);
}

Buffer * TLV :: GetBuffer(Int16 type)
{
  TLVElement * tlvElement = IsPresent(type);
  Buffer * buffer;

  if (!tlvElement)
    return (NULL);

  buffer = new Buffer(tlvElement->value, tlvElement->length, swap);

  return (buffer);
}

/****/

void TLV :: test()
{
  int i;

  cout << "TLV test() begin" << endl;

  for (i = 0; i < Length(); i ++) {
    cout << "    el # " << i << " is type " << elements[i].Type() <<
                                " is length " << elements[i].Length() << endl;
  }

  cout << "TLV test() done" << endl;
}
